from dataclasses import dataclass, asdict
from typing import Optional, Callable, Dict, Any

from lhotse.cut.text import TextExample
from lhotse.cut import Cut
from lhotse.dataset.sampling.base import (
    SamplingConstraint,
)
from lhotse.utils import is_none_or_gt

__all__ = ["TokenConstraint", "token_length_fn"]


def token_length_fn(cut: Cut) -> int:
    """
    A function to compute the number of tokens in a cut.
    This is used by the sampler to determine the length of each cut.
    """
    if isinstance(cut, TextExample):
        assert cut.tokens is not None
        return cut.num_tokens

    # get from feature
    num_tokens = cut.num_frames
    if num_tokens is None:
        # TODO(Feiteng): implement a more accuray way to get num_tokens for Audio & Image & Video Cuts
        # For now, we assume 100 frames per second, which is a common setting.
        num_tokens = int(cut.duration / 0.01)

    return num_tokens


@dataclass
class TokenConstraint(SamplingConstraint):
    """Represents a token-based constraint for sampler classes that sample text data.
    It is defined as maximum total number of tokens in a mini-batch and/or max batch size.

    Similarly to :class:`TimeConstraint`, we support ``quadratic_length`` for quadratic
    token penalty when sampling longer texts.
    """

    max_tokens: int = None
    max_cuts: int = None
    current: int = 0
    num_examples: int = 0
    longest_seen: int = 0
    token_length_fn: Callable[[Cut], int] = token_length_fn
    quadratic_length: Optional[int] = None

    def __post_init__(self) -> None:
        assert is_none_or_gt(self.max_tokens, 0)
        assert is_none_or_gt(self.max_cuts, 0)
        assert is_none_or_gt(self.quadratic_length, 0)

    def add(self, example: TextExample) -> None:
        """Increment the internal token counter for the constraint,
        selecting the right property from the input object.
        """
        if self.max_tokens is not None:
            size = self._maybe_apply_quadratic_correction(self.measure_length(example))
            self.current += size
            self.longest_seen = max(self.longest_seen, size)
        self.num_examples += 1

    def _maybe_apply_quadratic_correction(self, size: int) -> int:
        if self.quadratic_length is None:
            return size
        # For the quadratic complexity case, we add a term that accounts for
        # extra memory occupied by the model. The 1/quadratic_length term causes
        # the effective length to be doubled when it's equal to quadratic_length.
        return int(size + (size**2) / self.quadratic_length)

    def exceeded(self) -> bool:
        """Is the constraint exceeded or not."""
        if self.max_cuts is not None and self.num_examples > self.max_cuts:
            return True
        if self.max_tokens is None:
            return False
        effective_duration = self.num_examples * self.longest_seen
        return effective_duration > self.max_tokens

    def close_to_exceeding(self) -> bool:
        """Check if the batch is close to satisfying the constraints.
        We define "closeness" as: if we added one more cut that has
        duration/num_frames/num_samples equal to the longest seen cut
        in the current batch, then the batch would have exceeded the constraints.
        """
        if self.max_cuts is not None and self.num_examples >= self.max_cuts:
            return True

        if self.max_tokens is not None:
            effective_size = (self.num_examples + 1) * self.longest_seen
            return effective_size > self.max_tokens
        return False

    def reset(self) -> None:
        """Reset the internal counter (to be used after a batch was created,
        to start collecting a new one).
        """
        self.current = 0
        self.num_examples = 0
        self.longest_seen = 0

    def measure_length(self, example: TextExample) -> int:
        return self.token_length_fn(example)

    def state_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        for key, value in state_dict.items():
            setattr(self, key, value)



# modified from lhotse/audio/source.py

import io
import os
import warnings
from dataclasses import dataclass
from io import BytesIO, FileIO
from pathlib import Path
from subprocess import PIPE, run
from typing import List, Optional, Tuple, Union

import numpy as np
import soundfile as sf
import torch

from lhotse.audio.backend import read_audio
from lhotse.audio.utils import (
    DurationMismatchError,
    VideoInfo,
    VideoLoadingError,
    get_audio_duration_mismatch_tolerance,
)
from lhotse.caching import AudioCache
from lhotse.utils import (
    Pathlike,
    Seconds,
    SmartOpen,
    asdict_nonull,
    compute_num_samples,
    fastcopy,
)
from lhotse.audio.backend import SUPPORTED_VIDEO_EXTENSIONS, LibsndfileCompatibleAudioInfo


PathOrFilelike = Union[str, BytesIO, FileIO]


def torchaudio_ffmpeg_video_info(
    path_or_fileobj: Union[Path, str, BytesIO]
) -> VideoInfo:
    from torchaudio.io import StreamReader

    is_fileobj = not isinstance(path_or_fileobj, Path)
    is_mpeg = not is_fileobj and any(
        str(path_or_fileobj).endswith(ext)
        for ext in (".mp3",) + SUPPORTED_VIDEO_EXTENSIONS
    )
    if not is_fileobj:
        path_or_fileobj = str(path_or_fileobj)
    stream = StreamReader(path_or_fileobj)

    # Collect the information about available video and audio streams.
    num_streams = stream.num_src_streams
    audio_streams = {}
    video_streams = {}
    for stream_idx in range(num_streams):
        info = stream.get_src_stream_info(stream_idx)
        if info.media_type == "video":
            video_streams[stream_idx] = info
        elif info.media_type == "audio":
            audio_streams[stream_idx] = info
        else:
            raise RuntimeError(f"Unexpected media_type: {info}")

    assert (
        len(video_streams) < 2
    ), f"Lhotse currently does not support more than one video stream in a file (found {len(video_streams)})."
    assert len(audio_streams) < 2, (
        f"Lhotse currently does not support files with more than a single FFMPEG "
        f"audio stream yet (found {len(audio_streams)}). "
        f"Note that this is not the same as multi-channel which is generally supported."
    )

    meta = {}

    if video_streams:
        ((video_stream_idx, video_stream),) = list(video_streams.items())
        tot_frames = video_stream.num_frames

        if tot_frames == 0:  # num frames not available in header/metadata
            stream.add_basic_video_stream(
                round(video_stream.frame_rate), stream_index=video_stream_idx
            )
            for (chunk,) in stream.stream():
                tot_frames += chunk.shape[0]
            stream.remove_stream(0)

        meta["video"] = VideoInfo(
            fps=video_stream.frame_rate,
            height=video_stream.height,
            width=video_stream.width,
            num_frames=tot_frames,
        )

    if audio_streams:
        ((audio_stream_idx, audio_stream),) = list(audio_streams.items())
        stream.add_basic_audio_stream(
            frames_per_chunk=int(audio_stream.sample_rate),
            stream_index=audio_stream_idx,
        )

        def _try_read_num_samples():
            if is_mpeg or is_fileobj:
                # These cases often have insufficient or corrupted metadata, so we might need to scan
                # the full audio stream to learn the actual number of frames. If video is available,
                # we can quickly verify before performing the costly reading.
                video_info = meta.get("video", None)
                if video_info is not None:
                    audio_duration = audio_stream.num_frames / audio_stream.sample_rate
                    # for now 1ms tolerance
                    if abs(audio_duration - video_info.duration) < 1e-3:
                        return audio_stream.num_frames
                return 0
            else:
                return audio_stream.num_frames

        tot_samples = _try_read_num_samples()
        if tot_samples == 0:
            # There was a mismatch between video and audio duration in metadata,
            # we'll have to read the file to figure it out.
            for (chunk,) in stream.stream():
                tot_samples += chunk.shape[0]

        meta.update(
            channels=audio_stream.num_channels,
            frames=tot_samples,
            samplerate=int(audio_stream.sample_rate),
            duration=tot_samples / audio_stream.sample_rate,
        )
    else:
        # No audio stream in the video
        meta.update(
            channels=0,
            frames=0,
            samplerate=0,
            duration=meta["video"].duration,
        )

    return LibsndfileCompatibleAudioInfo(**meta)


@dataclass
class VideoSource:
    """
    VideoSource represents audio data that can be retrieved from somewhere.
    """

    type: str
    """
    The type of audio source. Supported types are:
    - 'file' (supports most standard audio encodings, possibly multi-channel)
    - 'command' [unix pipe] (supports most standard audio encodings, possibly multi-channel)
    - 'url' (any URL type that is supported by "smart_open" library, e.g. http/https/s3/gcp/azure/etc.)
    - 'memory' (any format, read from a binary string attached to 'source' member of VideoSource)
    - 'shar' (indicates a placeholder that will be filled later when using Lhotse Shar data format)
    """

    channels: List[int]
    """
    A list of integer channel IDs available in this VideoSource.
    """

    source: Union[str, bytes]
    """
    The actual source to read from. The contents depend on the ``type`` field,
    but in general it can be a path, a URL, or the encoded binary data itself.
    """

    video: Optional[VideoInfo] = None
    """
    Optional information about the video contained in this source, if any.
    """

    @property
    def has_video(self) -> bool:
        return self.video is not None

    @property
    def format(self) -> str:
        return self._get_format()

    def load_audio(
        self,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> np.ndarray:
        """
        Load the VideoSource (from files, commands, or URLs) with soundfile,
        accounting for many audio formats and multi-channel inputs.
        Returns numpy array with shapes: (n_samples,) for single-channel,
        (n_channels, n_samples) for multi-channel.

        Note: The elements in the returned array are in the range [-1.0, 1.0]
        and are of dtype `np.float32`.

        :param force_opus_sampling_rate: This parameter is only used when we detect an OPUS file.
            It will tell ffmpeg to resample OPUS to this sampling rate.
        """
        source = self._prepare_for_reading(offset=offset, duration=duration)

        samples, sampling_rate = read_audio(
            source,
            offset=offset,
            duration=duration,
            force_opus_sampling_rate=force_opus_sampling_rate,
        )

        # explicit sanity check for duration as soundfile does not complain here
        if duration is not None:
            num_samples = (
                samples.shape[0] if len(samples.shape) == 1 else samples.shape[1]
            )
            available_duration = num_samples / sampling_rate
            if (
                available_duration < duration - get_audio_duration_mismatch_tolerance()
            ):  # set the allowance as 1ms to avoid float error
                raise DurationMismatchError(
                    f"Requested more audio ({duration}s) than available ({available_duration}s)"
                )

        return samples.astype(np.float32)

    def load_video(
        self,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        with_audio: bool = True,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        import torchaudio

        try:

            # Open the video file for reading.
            stream = torchaudio.io.StreamReader(self.source)

            # Collect the information about available video and audio streams.
            num_streams = stream.num_src_streams
            audio_streams = {}
            video_streams = {}
            for stream_idx in range(num_streams):
                info = stream.get_src_stream_info(stream_idx)
                if info.media_type == "video":
                    video_streams[stream_idx] = info
                elif info.media_type == "audio":
                    audio_streams[stream_idx] = info
                else:
                    raise RuntimeError(f"Unexpected media_type: {info}")
            assert (
                len(video_streams) != 0
            ), "The file does not seem to have any video streams."
            assert (
                len(video_streams) < 2
            ), f"Lhotse currently does not support more than one video stream in a file (found {len(video_streams)})."
            assert len(audio_streams) < 2, (
                f"Lhotse currently does not support more than one audio stream in a file (found {len(video_streams)})."
                f" Note: it's different than multi-channel which is generally supported."
            )

            # Add an ffmpeg output video stream to perform reading in chunks.
            ((video_stream_idx, video_stream),) = list(video_streams.items())
            frames_per_chunk = round(video_stream.frame_rate)
            video_chunk_duration = frames_per_chunk * self.video.frame_length
            stream.add_basic_video_stream(
                frames_per_chunk,
                stream_index=video_stream_idx,
                height=self.video.height,  # re-scale if requested via overriding self.video.height
                width=self.video.width,  # re-scale if requested via overriding self.video.width
            )

            if with_audio and len(audio_streams) > 0:
                ((audio_stream_idx, audio_stream),) = list(audio_streams.items())
                samples_per_chunk = round(
                    audio_stream.sample_rate * video_chunk_duration
                )
                stream.add_basic_audio_stream(
                    samples_per_chunk,
                    stream_index=audio_stream_idx,
                )

            stream.seek(offset)
            video_chunks = []
            audio_chunks = []
            decoded_duration = 0.0
            for chunk in stream.stream():

                if duration is not None and decoded_duration >= duration:
                    break

                video_chunk = chunk[0]
                chunk_size = video_chunk.size(0)
                current_chunk_duration = chunk_size / video_stream.frame_rate

                if (
                    duration is not None
                    and decoded_duration + current_chunk_duration > duration
                ):
                    keep_frames = compute_num_samples(
                        video_chunk_duration - current_chunk_duration,
                        video_stream.frame_rate,
                    )
                    video_chunk = video_chunk[:keep_frames]

                video_chunks.append(video_chunk)

                if with_audio:
                    audio_chunk = chunk[1]
                    if (
                        duration is not None
                        and decoded_duration + current_chunk_duration > duration
                    ):
                        keep_samples = compute_num_samples(
                            video_chunk_duration - current_chunk_duration,
                            audio_stream.sample_rate,
                        )
                        audio_chunk = audio_chunk[:keep_samples]
                    audio_chunks.append(audio_chunk.T)

                decoded_duration += current_chunk_duration

            if not video_chunks:
                return (
                    torch.zeros(
                        0, 3, video_stream.height, video_stream.width, dtype=torch.uint8
                    ),
                    None,
                )

            output_video = torch.cat(video_chunks, dim=0)  # T x C x H x W
            output_audio = None
            if with_audio:
                output_audio = torch.cat(audio_chunks, dim=1)  # C x T

            return output_video, output_audio

        except Exception as e:
            raise VideoLoadingError(
                f"Reading video from '{self.source if not isinstance(self.source, bytes) else 'memory'}' failed. "
                f"Details: {type(e)}: {str(e)}"
            )

    def with_video_resolution(self, width: int, height: int) -> "VideoSource":
        return fastcopy(self, video=self.video.copy_with(width=width, height=height))

    def with_path_prefix(self, path: Pathlike) -> "VideoSource":
        if self.type != "file":
            return self
        return fastcopy(self, source=str(Path(path) / self.source))

    def to_dict(self) -> dict:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data) -> "VideoSource":
        if "video" in data:
            data["video"] = VideoInfo.from_dict(data["video"])
        return VideoSource(**data)

    def __repr__(self):
        return (
            f"VideoSource(type='{self.type}', channels={self.channels}, "
            f"source='{self.source if isinstance(self.source, str) else '<binary-data>'}')"
        )

    def _prepare_for_reading(
        self, offset: Seconds, duration: Optional[Seconds]
    ) -> PathOrFilelike:
        """
        Validates `self.type` and prepares the actual source for audio reading.
        Returns either a path or a file-like object opened in binary mode,
        that can be handled by :func:`lhotse.audio.backend.read_audio`.
        """
        assert self.type in (
            "file",
            "command",
            "url",
            "memory",
            "shar",
        ), f"Unexpected VideoSource type: '{self.type}'"

        source = self.source

        if self.type == "command":

            if (offset != 0.0 or duration is not None) and not AudioCache.enabled():
                warnings.warn(
                    "You requested a subset of a recording that is read from disk via a bash command. "
                    "Expect large I/O overhead if you are going to read many chunks like these, "
                    "since every time we will read the whole file rather than its subset."
                    "You can use `lhotse.set_caching_enabled(True)` to mitigate the overhead."
                )

            # Let's assume 'self.source' is a pipe-command with unchangeable file,
            # never a microphone-stream or a live-stream.
            audio_bytes = AudioCache.try_cache(self.source)
            if not audio_bytes:
                audio_bytes = run(self.source, shell=True, stdout=PIPE).stdout
                AudioCache.add_to_cache(self.source, audio_bytes)
            source = BytesIO(audio_bytes)

        elif self.type == "url":

            if offset != 0.0 or duration is not None and not AudioCache.enabled():
                warnings.warn(
                    "You requested a subset of a recording that is read from URL. "
                    "Expect large I/O overhead if you are going to read many chunks like these, "
                    "since every time we will download the whole file rather than its subset."
                    "You can use `lhotse.set_caching_enabled(True)` to mitigate the overhead."
                )

            # Let's assume 'self.source' is url to unchangeable file,
            # never a microphone-stream or a live-stream.
            audio_bytes = AudioCache.try_cache(self.source)
            if not audio_bytes:
                with SmartOpen.open(self.source, "rb") as f:
                    audio_bytes = f.read()
                AudioCache.add_to_cache(self.source, audio_bytes)
            source = BytesIO(audio_bytes)

        elif self.type == "memory":

            assert isinstance(self.source, bytes), (
                "Corrupted manifest: specified VideoSource type is 'memory', "
                f"but 'self.source' attribute is not of type 'bytes' (found: '{type(self.source).__name__}')."
            )
            source = BytesIO(self.source)

        elif self.type == "shar":

            raise RuntimeError(
                "Inconsistent state: found an VideoSource with Lhotse Shar placeholder "
                "that was not filled during deserialization."
            )

        return source

    def _get_format(self) -> str:
        """Get format for the audio source.
        If using 'file' or 'url' types, the format is inferred from the file extension, as in soundfile.
        If using 'memory' type, the format is inferred from the binary data.
        """
        if self.type in ("file", "url"):
            # Resolve audio format based on the filename
            format = os.path.splitext(self.source)[-1][1:]
            return format.lower()
        elif self.type == "memory":
            sf_info = sf.info(io.BytesIO(self.source))
            if sf_info.format == "OGG" and sf_info.subtype == "OPUS":
                # soundfile describes opus as ogg container with opus coding
                return "opus"
            else:
                return sf_info.format.lower()
        else:
            raise NotImplementedError(
                f"Getting format not implemented for source type {self.type}"
            )



"""Test the discoss package initialization."""

import discoss


def test_version():
    """Test that version is accessible."""
    assert hasattr(discoss, "__version__")
    assert isinstance(discoss.__version__, str)


def test_author():
    """Test that author information is accessible."""
    assert hasattr(discoss, "__author__")
    assert discoss.__author__ == "Feiteng Li"


def test_description():
    """Test that description is accessible."""
    assert hasattr(discoss, "__description__")
    assert "Distributed Coordinated Sequence Sampler" in discoss.__description__



"""DiscoSeqSampler - Distributed Coordinated Sequence Sampler."""


__version__ = "0.1.0"
__author__ = "Feiteng Li"
__email__ = ""
__description__ = "Distributed Coordinated Sequence Sampler"

# 在这里导入主要的类和函数
# from .sampler import DiscoSeqSampler
# from .coordinator import Coordinator

__all__: list[str] = [
    # "DiscoSeqSampler",
    # "Coordinator",
]



from .constraint import TokenConstraint
from .simple import SimpleCutSampler
from .dynamic_bucketing import DynamicBucketingCutSampler
from .bucketing import BucketingCutSampler


__all__ = [
    "TokenConstraint",
    "SimpleCutSampler",
    "BucketingCutSampler",
    "DynamicBucketingCutSampler",
]



import warnings
from typing import Any, Dict, Optional

from lhotse import CutSet
from lhotse.dataset.sampling import SimpleCutSampler as LhotseSimpleCutSampler

from .constraint import TokenConstraint

__all__ = ["SimpleCutSampler"]


class SimpleCutSampler(LhotseSimpleCutSampler):
    """A simple cut sampler that inherits from Lhotse's SimpleCutSampler."""

    def __init__(
        self,
        cuts: CutSet,
        constraint: TokenConstraint,
        shuffle: bool = False,
        drop_last: bool = False,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        seed: int = 0,
    ):
        super().__init__(
            cuts=cuts,
            max_cuts=1000000000,  # No limit on the number of cuts
            shuffle=shuffle,
            drop_last=drop_last,
            world_size=world_size,
            rank=rank,
            seed=seed,
        )
        self.time_constraint = constraint

    def state_dict(self) -> Dict[str, Any]:
        state_dict = super().state_dict()
        state_dict.update(
            {
                "token_constraint": self.time_constraint.state_dict(),
            }
        )
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        time_constraint = TokenConstraint(**state_dict.pop("token_constraint"))
        if self.time_constraint != time_constraint:
            warnings.warn(
                "SimpleCutSampler.load_state_dict(): Inconsistent token_constraint:\n"
                f"expected {self.time_constraint}\n"
                f"received {time_constraint}\n"
                f"We will overwrite the settings with the received state_dict."
            )
        self.time_constraint = time_constraint

        super().load_state_dict(state_dict)



from lhotse.dataset.sampling import (
    DynamicBucketingSampler as LhotseDynamicBucketingSampler,
)
from dataclasses import asdict, dataclass
from typing import (
    Iterable,
    List,
    Literal,
    Optional,
    Union,
)

from lhotse import Seconds
from .constraint import TokenConstraint

__all__ = ["DynamicBucketingCutSampler"]


class DynamicBucketingCutSampler(LhotseDynamicBucketingSampler):
    def __init__(
        self,
        *cuts: Iterable,
        constraint: TokenConstraint,
        num_buckets: Optional[int] = 10,
        shuffle: bool = False,
        drop_last: bool = False,
        consistent_ids: bool = True,
        bucket_bins: List[int] = None,
        num_cuts_for_bins_estimate: int = 10000,
        buffer_size: int = 20000,
        quadratic_duration: Optional[Seconds] = None,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        seed: Union[int, Literal["randomized", "trng"]] = 0,
        sync_buckets: bool = True,
        concurrent: bool = False,
        strict=None,
    ):
        super().__init__(
            *cuts,
            constraint=constraint,
            num_buckets=num_buckets,
            shuffle=shuffle,
            drop_last=drop_last,
            consistent_ids=consistent_ids,
            duration_bins=bucket_bins,
            num_cuts_for_bins_estimate=num_cuts_for_bins_estimate,
            buffer_size=buffer_size,
            quadratic_duration=quadratic_duration,
            world_size=world_size,
            rank=rank,
            seed=seed,
            sync_buckets=sync_buckets,
            concurrent=concurrent,
            strict=strict,
        )



#!/usr/bin/env python

"""Setup script for DiscoSeqSampler."""

import os

from setuptools import find_packages, setup


# Read the README file
def read_readme():
    """Read README.md file."""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        return f.read()


# Read version from __init__.py
def get_version():
    """Get version from discoss/__init__.py."""
    here = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(here, "discoss", "__init__.py")

    # If __init__.py doesn't have version, return default
    if not os.path.exists(version_file):
        return "0.1.0"

    with open(version_file, encoding="utf-8") as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("=")[1].strip().strip('"').strip("'")

    return "0.1.0"


setup(
    name="discoss",
    version=get_version(),
    description="Distributed Coordinated Sequence Sampler",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    author="Feiteng Li",
    author_email="lifeiteng0422@gmail.com",
    url="https://github.com/lifeiteng/DiscoSeqSampler",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.10",
    install_requires=[
        "lhotse==1.30.3",
        "torch==2.7.1",
        "torchaudio==2.7.1",
        "torchcodec==0.5",
    ],
    extras_require={
        "dev": [
            # Testing
            "pytest>=7.0",
            "pytest-cov>=4.0",
            "pytest-xdist>=3.0",
            # Code formatting and linting
            "black>=23.0",
            "isort>=5.12",
            "ruff>=0.1.0",
            "mypy>=1.5",
            # Pre-commit hooks
            "pre-commit>=3.0",
            # Documentation
            "sphinx>=5.0",
            "sphinx-rtd-theme>=1.0",
        ],
        "docs": [
            "sphinx>=5.0",
            "sphinx-rtd-theme>=1.0",
            "myst-parser>=2.0",
        ],
    },
    entry_points={
        "console_scripts": [
            # "discoss=discoss.cli:main",
        ],
    },
    include_package_data=True,
    zip_safe=False,
    keywords="distributed sampling sequence coordination",
)



"""Test the discoss package initialization."""

from lhotse.cut import CutSet
from discoss.sampler import BucketingCutSampler, TokenConstraint
import logging


def _test_sampler(
    cuts_file,
    rank,
    world_size,
    constraint: TokenConstraint = None,
    max_cuts: int = 10,
    drop_last: bool = False,
):
    """Test that sampler is accessible and data items."""
    cuts = CutSet.from_file(cuts_file)
    sampler = BucketingCutSampler(
        cuts,
        constraint=constraint
        or TokenConstraint(
            max_tokens=4000,
            max_cuts=max_cuts,
            quadratic_length=None,
        ),
        num_buckets=10,
        drop_last=drop_last,
        shuffle=True,
        rank=rank,
        world_size=world_size,
    )

    batches, num_cuts = [], 0
    for batch in sampler:
        assert batch is not None
        num_cuts += len(batch)
        batches.append(batch)

    assert len(cuts) // world_size - num_cuts <= (
        max_cuts * (int(drop_last) + 1)
    ), f"{rank=} {world_size=} Expected {len(cuts) // world_size} cuts, got {num_cuts} cuts({max_cuts=})."

    # num_batches = len(batches)
    # logging.info(
    #     f"{rank=} {world_size=}: Number of cuts sampled: {num_cuts} vs {len(cuts) // world_size} expected({num_batches=})."
    # )
    return batches


def test_sampler_dynamic_bucketing_gpu1(cuts_file):
    """Test that sampler is accessible."""
    _test_sampler(cuts_file, 0, 1)


def test_sampler_dynamic_bucketing_gpu2(cuts_file):
    """Test that sampler is accessible."""
    _test_sampler(cuts_file, 0, 2)
    _test_sampler(cuts_file, 1, 2)


def test_sampler_dynamic_bucketing_gpu4(cuts_file):
    """Test that sampler is accessible."""
    _test_sampler(cuts_file, 2, 4)
    _test_sampler(cuts_file, 3, 4)


def test_sampler_dynamic_bucketing_gpu8(cuts_file):
    """Test that sampler is accessible."""
    _test_sampler(cuts_file, 0, 8)
    _test_sampler(cuts_file, 1, 8)
    _test_sampler(cuts_file, 2, 8)
    _test_sampler(cuts_file, 3, 8)
    _test_sampler(cuts_file, 4, 8)
    _test_sampler(cuts_file, 5, 8)
    _test_sampler(cuts_file, 6, 8)
    _test_sampler(cuts_file, 7, 8)


def test_sampler_dynamic_bucketing_gpu4_data_sanity(cuts_file):
    max_tokens, max_cuts = 4000, 10
    constraint = TokenConstraint(
        max_tokens=max_tokens,
        max_cuts=max_cuts,
        quadratic_length=None,
    )

    batches_1 = _test_sampler(cuts_file, 0, 4, constraint=constraint, drop_last=True)
    batches_2 = _test_sampler(cuts_file, 1, 4, constraint=constraint, drop_last=True)
    batches_3 = _test_sampler(cuts_file, 2, 4, constraint=constraint, drop_last=True)
    batches_4 = _test_sampler(cuts_file, 3, 4, constraint=constraint, drop_last=True)

    # 1. check num batches(drop_last=True)
    assert len(batches_1) == len(batches_2) == len(batches_3) == len(batches_4)

    bad_count = 0
    for b1, b2, b3, b4 in zip(batches_1, batches_2, batches_3, batches_4):
        batch_size = [len(b1), len(b2), len(b3), len(b4)]

        # 2. check max_cuts
        assert max(batch_size) <= max_cuts
        if max(batch_size) - min(batch_size) >= 2:
            bad_count += 1

        # 3. check data uniqueness
        # Ensure that cuts in each batch are unique across all batches
        cut_ids = (
            [c.id for c in b1]
            + [c.id for c in b2]
            + [c.id for c in b3]
            + [c.id for c in b4]
        )
        assert len(cut_ids) == len(
            set(cut_ids)
        ), "Cut IDs are not unique across batches."

        # 4. check num tokens
        for b in [b1, b2, b3, b4]:
            lengths = [constraint.measure_length(cut) for cut in b]
            if max(lengths) <= max_tokens:
                assert sum(lengths) <= max_tokens + max(lengths)
            else:
                logging.info(f"{lengths=} sum > {max_tokens=}")

    # 5. check batch size consistency across ranks
    assert (
        bad_count / len(batches_1) < 0.1
    ), f"Batch sizes are not consistent across ranks: {bad_count}/{len(batches_1)}."



"""Test the discoss package initialization."""

from lhotse.cut import CutSet
from discoss.sampler import SimpleCutSampler, TokenConstraint
import logging


def _test_sampler(cuts, rank, world_size, max_cuts: int = 10):
    """Test that sampler is accessible and data items."""
    sampler = SimpleCutSampler(
        cuts,
        constraint=TokenConstraint(
            max_tokens=1000,
            max_cuts=max_cuts,
            quadratic_length=None,
        ),
        rank=rank,
        world_size=world_size,
    )

    num_cuts = 0
    for batch in sampler:
        assert batch is not None
        num_cuts += len(batch)

    assert (
        len(cuts) // world_size - num_cuts <= max_cuts
    ), f"{rank=} {world_size=} Expected {len(cuts) // world_size} cuts, got {num_cuts} cuts({max_cuts=})."
    # logging.info(f"{rank=} {world_size=}: Number of cuts sampled: {num_cuts} vs {len(cuts) // world_size} expected.")


def test_sampler_simple_gpu1(audio_cuts):
    """Test that sampler is accessible."""
    _test_sampler(audio_cuts, 0, 1)


def test_sampler_simple_gpu2(audio_cuts):
    """Test that sampler is accessible."""
    _test_sampler(audio_cuts, 0, 2)
    _test_sampler(audio_cuts, 1, 2)


def test_sampler_simple_gpu4(audio_cuts):
    """Test that sampler is accessible."""
    _test_sampler(audio_cuts, 2, 4)
    _test_sampler(audio_cuts, 3, 4)


def test_sampler_simple_gpu8(audio_cuts):
    """Test that sampler is accessible."""
    _test_sampler(audio_cuts, 0, 8)
    _test_sampler(audio_cuts, 1, 8)
    _test_sampler(audio_cuts, 2, 8)
    _test_sampler(audio_cuts, 3, 8)
    _test_sampler(audio_cuts, 4, 8)
    _test_sampler(audio_cuts, 5, 8)
    _test_sampler(audio_cuts, 6, 8)
    _test_sampler(audio_cuts, 7, 8)
